#ifndef __ayyi_client_h__
#define __ayyi_client_h__

#include "ayyi/ayyi_log.h"
#include "ayyi/ayyi_typedefs.h"
#include "ayyi/ayyi_types.h"

#define ASSERT_SONG_SHM  if(!ayyi.got_shm){ errprintf2(__func__, "no song shm.\n"); return; }
#define ASSERT_SONG_SHM_RET_FALSE  if(!ayyi.got_shm){ errprintf2(__func__, "no song shm.\n"); return FALSE; }
#define UNEXPECTED_PROPERTY(T) gwarn("unexpected property: obj_type=%s prop=%s", ayyi_print_object_type(T), ayyi_print_prop_type(prop));
#define UNEXPECTED_OBJECT_TYPE(T) gwarn("unexpected object type: %i %s", T, ayyi_print_object_type(T));

#define AYYI_NULL_IDENT ((AyyiIdent){-1, -1})

#ifndef __ayyi_client_c__
extern AyyiService* known_services[];
#endif

#if (defined __ayyi_client_c__ || defined __ayyi_action_c__ || defined __ayyi_shm_c__ || defined __am_message_c__ )
struct _AyyiClientPrivate
{
	GSList*                 plugins;    // loaded Ayyi plugins

	GList*                  actionlist; // list of AyyiAction*
	unsigned                trans_id;

	gboolean                (*on_shm)(struct _shm_seg*); //used for segment verification.
	//gboolean                (*on_shm)(AyyiShmCallbackData*);

	GHashTable*             responders;
};
#endif

#if (defined __ayyi_dbus_c__ || defined __ayyi_client_c__ || defined __dbus_proxy_c__ || defined __ayyi_shm_c__ )
#include <dbus/dbus-glib-bindings.h>
struct _AyyiServicePrivate
{
	DBusGProxy*             proxy;

	void                    (*on_shm)(AyyiShmCallbackData*); // application level callback
};
#endif

typedef struct _AyyiClientPrivate AyyiClientPrivate;
typedef struct _AyyiServicePrivate AyyiServicePrivate;

struct _ayyi_client
{
	AyyiService*            service;     // deprecated - use the array below instead.
	AyyiService**           services;    // null terminated.

	int                     got_shm;     // TRUE when we must have _all_ required shm pointers.
	gboolean                offline;     // TRUE if we are not connected and should not attempt to connect to remote service.

	int                     signal_mask; // defines which signals we wish to receive.

	struct _log             log;

	GList*                  launch_info; // type AyyiLaunchInfo

	AyyiClientPrivate*      priv;
};

struct _service
{
	char*                   service;
	char*                   path;
	char*                   interface;

	//instance values (service is singleton, so class and instance are the same):

	void                    (*ping)      (void (*)(const char*));

	GList*                  segs;        // list of type shm_seg*

	AyyiServicePrivate*     priv;
};

typedef enum {
	AYYI_SERVICE_AYYID1 = 0,
	AYYI_SERVICE_MIXER,
} AyyiServiceType;

typedef struct plugin {
	guint       api_version;
	gchar*      name;
	guint       type;          // plugin type (currently only PLUGIN_TYPE_1)

	gchar*      service_name;
	gchar*      app_path;
	gchar*      interface;

	void*       symbols;       // plugin type specific symbol table

	void*       client_data;   // currently used to store the local shm address;
} *AyyiPluginPtr;

typedef struct _responder
{
	AyyiHandler callback;
	gpointer    user_data;
} Responder;

struct _shm_callback_data
{
	shm_seg*     shm_seg;
	gpointer     user_data;
	AyyiService* service;
};


AyyiClient*  ayyi_client_init                 ();
void         ayyi_client_add_services         (AyyiService**);
gboolean     ayyi_client_connect              (AyyiService*, AyyiHandler2, gpointer, GError** error);
void         ayyi_client_load_plugins         ();
AyyiPluginPtr ayyi_client_get_plugin          (const char*);
void         ayyi_client_set_signal_reg       (int);

void         ayyi_object_set_bool             (AyyiObjType, AyyiIdx, int prop, gboolean val, AyyiAction*);
gboolean     ayyi_object_is_deleted           (AyyiRegionBase*);

void         ayyi_client_add_responder        (AyyiIdent, AyyiOpType, AyyiHandler, gpointer);
void         ayyi_client_add_onetime_responder(AyyiIdent, AyyiOpType, AyyiHandler, gpointer);
const GList* ayyi_client_get_responders       (AyyiObjType, AyyiOpType, AyyiIdx);
gboolean     ayyi_client_remove_responder     (AyyiObjType, AyyiOpType, AyyiIdx, AyyiHandler);
void         ayyi_client_print_responders     ();

void         ayyi_discover_clients            ();

AyyiIdx      ayyi_ident_get_idx               (AyyiIdent);
AyyiObjType  ayyi_ident_get_type              (AyyiIdent);

void         ayyi_launch                      (AyyiLaunchInfo*, gchar** argv);
void         ayyi_launch_by_name              (const char*, gchar** argv);

gboolean     ayyi_verify_playlists            ();

const char*  ayyi_print_object_type           (AyyiObjType);
const char*  ayyi_print_optype                (AyyiOpType);
const char*  ayyi_print_prop_type             (AyyiPropType);
const char*  ayyi_print_media_type            (AyyiMediaType);

#ifdef __ayyi_client_c__
struct _ayyi_client ayyi;
#else
extern struct _ayyi_client ayyi;
#endif

#endif //__ayyi_client_h__
