#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <jack/jack.h>
#include <glib.h>
#include <ayyi/ayyi_typedefs.h>
#include <ayyi/ayyi_types.h>
#include <ayyi/ayyi_utils.h>
#include <ayyi/interface.h>
#include <ayyi/ayyi_msg.h>
#include <ayyi/ayyi_time.h>
#include <ayyi/ayyi_client.h>
#include <services/ardourd.h>

extern struct _ayyi_client ayyi;

static const uint32_t ticks_per_beat = 4 * TICKS_PER_SUBBEAT;


void
ayyi_pos2bbst(AyyiSongPos* pos, char* str)
{
	//converts a song position to a display string in bars, beats, subbeats, ticks.

	sprintf(str, "%03i:%02i:%02i:%03i",
						pos->beat / 4,
						pos->beat % 4,
						pos->sub,
						pos->mu);
}


int64_t
ayyi_pos2mu(AyyiSongPos* pos)
{
	return ((int64_t)pos->mu) + (pos->sub * AYYI_MU_PER_SUB) + (((int64_t)pos->beat) * AYYI_MU_PER_SUB*AYYI_SUBS_PER_BEAT);
}


void
ayyi_pos_add_mu(AyyiSongPos* pos, uint64_t mu)
{
	mu += pos->mu;
	unsigned long subs = mu / AYYI_MU_PER_SUB; //carry overflow into the subs field.
	pos->mu            = mu % AYYI_MU_PER_SUB; //use the remainder.
	dbg (1, "mu=%Li remainder=%Li", mu, mu % AYYI_MU_PER_SUB);

	subs += pos->sub;
	long beats         = subs / AYYI_SUBS_PER_BEAT;
	pos->sub           = subs % AYYI_SUBS_PER_BEAT;

	dbg (1, "overflow=%li pos->beat=%i", beats, pos->beat);
	pos->beat += beats; //FIXME too high

	dbg (1, "%i:%i:%i", pos->beat, pos->sub, pos->mu);
}


gboolean
ayyi_pos_is_valid(AyyiSongPos* pos)
{
	if(!pos) return false;
	if(pos->sub >= AYYI_SUBS_PER_BEAT) return false;
	if(pos->mu >= AYYI_MU_PER_SUB) return false;
	return TRUE;
}


void
ayyi_samples2pos(uint32_t samples, AyyiSongPos* pos)
{
	int64_t mu = ayyi_samples2mu(samples);
	ayyi_mu2pos(mu, pos);
}


uint32_t
ayyi_pos2samples(AyyiSongPos* pos)
{
	float bpm = ((AyyiSongService*)ayyi.service)->song->bpm;

	float length_of_1_beat_in_secs = 60.0/bpm;
	//float mu = pos->mu;
	float b = ((float)AYYI_SUBS_PER_BEAT) * ((float)AYYI_MU_PER_SUB);
	float beat = pos->beat;
	float sub = pos->sub;
	//dbg(0, "%.2f %.2f %.2f", beat, sub, mu);
	float samples = length_of_1_beat_in_secs * ((float)((AyyiSongService*)ayyi.service)->song->sample_rate) *
		(
			beat +
			sub / ((float)AYYI_SUBS_PER_BEAT) +
			((float)pos->mu) / b
		);

	return (uint32_t)(samples + 0.5);
}


long long 
ayyi_samples2mu(unsigned long long samples)
{
	//returns the length in mu, of the given number of audio samples.

	//normally, the ratio is of the order of 1924

	long long mu;

	int sample_rate = ((AyyiSongService*)ayyi.service)->song->sample_rate;
	if(!sample_rate){ gerr ("samplerate zero!"); sample_rate=44100; }
	//dbg(0, "samplerate=%i.\n", sample_rate);

	float secs = (float)samples / (float)sample_rate; //length in seconds.

	float bpm = ((AyyiSongService*)ayyi.service)->song->bpm;
	float length_of_1_beat_in_secs = 60.0/bpm;

	float beats = secs / length_of_1_beat_in_secs;

	//float remainder = secs - beats * 

	mu = ayyi_beats2mu(beats);
	//dbg(0, "float=%.3fsecs 1beat=%.3fsecs beats=%.3f mu=%Li\n", secs, length_of_1_beat_in_secs, beats, mu);

	//FIXME this check below indicates there are some largish rounding errors:
	dbg (2, "samples=%Li mu=%Li check=%Li.", samples, mu, ayyi_mu2samples(mu));
	return mu;
}


long long
ayyi_mu2samples(long long mu)
{
	float mu_f = (float)mu;
	float beats = mu_f / (AYYI_SUBS_PER_BEAT * AYYI_MU_PER_SUB);

	dbg(3, "mu=%Li (%.4f) beats=%.4f", mu, mu_f, beats);

	return ayyi_beats2samples_float(beats);
}


void
ayyi_mu2pos(uint64_t mu, AyyiSongPos* pos)
{
	g_return_if_fail(pos);

	uint32_t beats = mu / (AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT);

	uint32_t remainder  = mu % (AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT);
	uint32_t subs = remainder / AYYI_MU_PER_SUB;
	remainder = remainder % AYYI_MU_PER_SUB;

	pos->beat = beats;
	pos->sub  = subs;
	pos->mu   = remainder;
}


#if 0
void
ayyi_pos_set(AyyiSongPos* pos, int beat, uint16_t sub, uint16_t mu)
{
	g_return_if_fail(pos);

	pos->beat = beat;
	pos->sub  = sub;
	pos->mu   = mu;
}
#endif


uint64_t
ayyi_beats2mu(double beats)
{
	//note beats is float.

	return beats * (AYYI_MU_PER_SUB * AYYI_SUBS_PER_BEAT); //this is the definition of a mu.
}


long long
ayyi_beats2samples(int beats)
{
	//returns the number of audio samples equivalent to the given number of beats.

	float bpm = ((AyyiSongService*)ayyi.service)->song->bpm;

	float length_of_1_beat_in_secs = 60.0/bpm;
	float samples = length_of_1_beat_in_secs * ((AyyiSongService*)ayyi.service)->song->sample_rate * beats;

	//dbg(0, "%ibeats = %.2fsamples. length_of_1_beat_in_secs=%.2f", beats, samples, length_of_1_beat_in_secs);

	return (int)samples;
}


long long
ayyi_beats2samples_float(float beats)
{
	//returns the number of audio samples equivalent to the given number of beats.

	float bpm = ((AyyiSongService*)ayyi.service)->song->bpm;

	float length_of_1_beat_in_secs = 60.0 / bpm;
	float samples = length_of_1_beat_in_secs * ((AyyiSongService*)ayyi.service)->song->sample_rate * beats;

	dbg (3, "%.4fbeats = %.4fsamples. length_of_1_beat_in_secs=%.4f", beats, samples, length_of_1_beat_in_secs);

	return (int)(samples + 0.5);
}


gboolean
ayyi_pos_cmp(const AyyiSongPos* a, const AyyiSongPos* b)
{
	//return TRUE if the two positions are different.
	return (a->beat != b->beat) || (a->sub != b->sub) || (a->mu != b->mu);
}


void
ayyi_pos_add(AyyiSongPos* pos, const AyyiSongPos* addition)
{
	int64_t mu = ayyi_pos2mu(pos) + ayyi_pos2mu(addition);
	ayyi_mu2pos(mu, pos);
}


void
ayyi_samples2bbst(uint64_t samples, char* str)
{
	if(_debug_>-1) if(samples > ayyi_beats2samples(AYYI_MAX_BEATS)){ gwarn ("samples too high? %Lu", samples); str[0]='\0'; return; }

	struct _shm_song* song = ((AyyiSongService*)ayyi.service)->song;

	float beats_per_second = song->bpm / 60.0;

	float secs = (float)samples / (float)((AyyiSongService*)ayyi.service)->song->sample_rate;
	int ticks = secs * beats_per_second * ticks_per_beat;

	float beats2bar = 4.0;
	int bar = ticks / (ticks_per_beat * (int)beats2bar);
	int ticks_remaining = ticks % (ticks_per_beat * (int)beats2bar);

	int beat = 0;
	if     (ticks_remaining < 1*ticks_per_beat) beat = 0;
	else if(ticks_remaining < 2*ticks_per_beat) beat = 1;
	else if(ticks_remaining < 3*ticks_per_beat) beat = 2;
	else if(ticks_remaining < 4*ticks_per_beat) beat = 3;
	ticks_remaining = ticks_remaining - beat * ticks_per_beat;

	int subbeat = ticks_remaining / TICKS_PER_SUBBEAT;
	ticks_remaining = ticks_remaining % TICKS_PER_SUBBEAT;

	//float seconds_remaining = secs - (subbeat2 * beats_per_second / 4);
	//int tick = seconds_remaining * 3840 / beats_per_second;

	sprintf(str, BBST_FORMAT, bar, beat+1, subbeat+1, ticks_remaining);
	//dbg(0, "FIXME samples=%Lu %.3fs ticks=%i sub=%i %s %i/4\n", samples, secs, ticks, subbeat, str, (int)beats2bar); 
}


