/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2004-2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <math.h>
#include <gtk/gtk.h>

#include <model/common.h>
#include <ayyi/ayyi_list.h>
#include <ayyi/ayyi_shm.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_mixer.h>
#include <model/am_plugin.h>
#include <model/am_song.h>
#include "model/am_mixer.h"
#include "model/am_connections.h"
#include "model/am_track.h"
#include "model/am_collection.h"
#include "model/am_channel.h"

#define g_object_unref0(var) ((var == NULL) ? NULL : (var = (g_object_unref (var), NULL)))

extern GHashTable* plugin_icons;

static GtkTreeModel* am_channel_create_plugin_model();
static void          channel_init(Channel*, AyyiIdx);
static void          channel_free(Channel*);

static GdkPixbuf* blank_icon = NULL;

//note: the master channel no longer has to be at the end of the channel array.

AMArray*
channels__init(int n_channels)
{
	GBoxedCopyFunc t_dup_func = NULL;
	GDestroyNotify t_destroy_func = NULL;
	GType t_type = 0; //FIXME --------------------------------------------------------------------------------------- <<<<<<
	song->channels = am_array_new (t_type, t_dup_func, t_destroy_func, n_channels);

#if 0     // cannot do this here - stock items not declared!
	GtkIconSet* icon_set = gtk_icon_factory_lookup_default(SM_STOCK_BLANK);
	if(icon_set){
		blank_icon = gtk_icon_set_render_icon(icon_set, /*style*/NULL, GTK_TEXT_DIR_NONE, GTK_STATE_NORMAL, 16, NULL, NULL);
	}
#else
	blank_icon = blank_icon ? blank_icon : gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, 16, 16);
	gdk_pixbuf_fill(blank_icon, 0);
#endif

	GPtrArray* channels = song->channels->array;
	int i;
	for(i=0;i<channels->len;i++){
		Channel* channel = g_new0(Channel, 1); //g_ptr_array_index(song->channels, i);
		g_ptr_array_add(channels, channel);
		channel_init(channel, -1);
	}

	return song->channels;
}


void
channels__load()
{
	// sync the channel-list to the core.
	// Used primarily following a new song load.

	if(!((AyyiSongService*)ayyi.service)->amixer) return;

	Channel* channel = NULL;

	int n_channels = ayyi_mixer__count_channels();
	int old_chan_list_len = 0;

	//allocate the array:
	if (!song->channels){
		dbg (0, "creating new channel list... tot_strips=%i", n_channels);
		channels__init(n_channels);
	} else {
		old_chan_list_len = song->channels->array->len;
		dbg (2, "channel list size: %i --> %i.", old_chan_list_len, n_channels);
		if(old_chan_list_len < n_channels){
			int i; for(i=0;i<n_channels-old_chan_list_len;i++){
				g_ptr_array_add(song->channels->array, g_new0(Channel, 1));
			}
		}
		else if(old_chan_list_len > n_channels){
			dbg (2, "shortening the array...");

			//remove references to this strip before we delete it:
			int i; for(i=ayyi_mixer__count_channels();i<old_chan_list_len;i++){
				Channel* channel = g_ptr_array_index(song->channels->array, i);
				if(channel->type == CHAN_MASTER) continue;

				am_channels__emit("delete", channel);
			}

			g_ptr_array_set_size(song->channels->array, n_channels);
			dbg (2, "real new size=%i", song->channels->array->len);
		}
	}

	channel = NULL;
	int i;
	for(i=0;i<n_channels;i++){
		int core_idx = am__mixer_channel_next(channel);
		AyyiTrack* r = ayyi_song__audio_track_at(core_idx);
		AyyiChannel* ch = ayyi_mixer__channel_at(core_idx);
		dbg (2, "%i: setting pod_idx=%i", i, core_idx);

		channel = g_ptr_array_index(song->channels->array, i);
		dbg (2, "channel[%i]=%p", i, channel);

		channel->type        = AYYI_TRACK_IS_MASTER(r) ? CHAN_MASTER : CHAN_INPUT;
		channel->nchans      = ch->nchans;//(channel->type == CHAN_MASTER) ? 2 : 1;
		channel_init(channel, core_idx);
	}

	am_channels__emit("change");

	dbg (2, "tot_strips=%i", n_channels);
}


void
channels__append_new(TrackNum tnum)
{
	Channel* new_channel = g_new0(Channel, 1);
	GPtrArray* array = song->channels->array;
	int ch_num = array->len;
	g_ptr_array_add(array, new_channel);

	Channel* si     = (Channel*)am_array_at(song->channels, ch_num);
	si->type        = CHAN_INPUT;
	si->nchans      = 1;
	channel_init(si, song->tracks->track[tnum]->ident.idx);

	dbg(2, "ch_num=%i type=%i new_n_channels=%i", ch_num, si->type, array->len);
#if 0
	channels_print();
#endif

	am_channels__emit ("add", new_channel);
}


void
channels__remove_channel(AMTrack* tr)
{
	//remove the given track from the channel list.
	PF2;
	GPtrArray* array = song->channels->array;
	int old_len = array->len;

	int s = am_track__lookup_channel_num(tr);
	if(s > -1){
		dbg (2, "chan=%i", s);
		Channel* old_channel = g_ptr_array_index(array, s);

		//note: mixer_win_delete_strip() requires that we modify the channel array *after* the mixer strips have been modified.
        am_channels__emit("delete", old_channel);

		channel_free(old_channel);
		if(!g_ptr_array_remove_index(array, s)) gerr("!!");
		dbg (2, "len: %i --> %i", old_len, array->len);
		if(s < array->len) dbg(2, "type=%i", ((Channel*)g_ptr_array_index(array, s))->type);
	}
	//am_song__channels_print();
}


Channel*
channels__get_master()
{
	GPtrArray* array = song->channels->array;
	int i; for(i=0;i<array->len;i++){
		Channel* channel = g_ptr_array_index(array, i);
		if(channel->type == CHAN_MASTER) return channel;
	}
	gwarn("master not found.");
	return NULL;
}


Channel*
channels__get_by_index(AyyiIdx idx)
{
	GPtrArray* array = song->channels->array;
	int i; for(i=0;i<array->len;i++){
		Channel* channel = g_ptr_array_index(array, i);
		if(channel->core_index == idx) return channel;
	}
	gwarn("channel not found.");
	return NULL;
}


static void
channel_init(Channel* channel, AyyiIdx idx)
{
	//note: because channel allocation is done as part of the list, its difficult to have a separate channel_new() function.
	//      Make sure all channel properties are properly initialised here

	channel->core_index  = idx;

	if(!channel->plugin_list) channel->plugin_list = am_channel_create_plugin_model();
	if(!channel->aux_level[0]){
		int a; for(a=0;a<AYYI_AUX_PER_CHANNEL;a++) channel->aux_level[a] = GTK_OBJECT(gtk_adjustment_new (0, -100.0, 11.0, 0.01, 0.1, 0));
	}

	if(!channel->meter_level){
		//float meter_min = -70.0; // range of gtk_meter
		float meter_min = -96.0;   // range of bst_meter
		float meter_max =   0.0;
		channel->meter_level = GTK_OBJECT(gtk_adjustment_new (0, meter_min, meter_max, 0.01, 0.1, 0.1));
	}
}


static void
channel_free(Channel* channel)
{
	if(g_object_is_floating(channel->meter_level)) g_object_ref_sink(channel->meter_level);
	g_object_unref0(channel->meter_level);

	g_object_unref(channel->plugin_list);
	g_free(channel);
}


#if 0
Channel*
channel_first(AMIter* iter)
{
#warning channel_first removeme
	iter->i = -1;
	g_return_val_if_fail(song->channels, NULL);

	iter->i = 0;
	GPtrArray* array = song->channels->array;
	Channel* channel = g_ptr_array_index(array, 0);
	return channel;
}


Channel*
c hannel_next(AMIter* iter)
{
	GPtrArray* array = song->channels->array;
	if(iter->i >= array->len || iter < 0){ return NULL; }
	Channel* channel = g_ptr_array_index(array, iter->i++);
	return channel;
}
#endif


bool
am_channel__is_muted(Channel* channel)
{
	g_return_val_if_fail(channel, false);

	AyyiTrack* track = ayyi_song__audio_track_at(channel->core_index);
	return track ? track->flags & muted : FALSE;
}


bool
am_channel__is_solod(Channel* channel)
{
	g_return_val_if_fail(channel, false);

	AyyiTrack* track = ayyi_song__audio_track_at(channel->core_index);
	if(track){
		return track->flags & solod;
	}
	return FALSE;
}


bool
am_channel__is_sdef(Channel* channel)
{
	return FALSE;
}


bool
am_channel__is_armed(Channel* channel)
{
	g_return_val_if_fail(channel, false);

	AyyiTrack* track = ayyi_song__audio_track_at(channel->core_index);
	return track ? track->flags & armed : FALSE;
}


const char*
am_channel__get_name(Channel* ch)
{
	AyyiChannel* ac = ayyi_mixer__channel_at(ch->core_index);
	return ac ? ac->name : NULL;
}


double
am_channel__fader_level(Channel* channel)
{
	AyyiChannel* ch = ayyi_mixer__channel_at_quiet(channel->core_index);
	g_return_val_if_fail(ch, 0.0);

	//static int ddd = 0;
	//if (channel->core_index == 1 && !(ddd++%10)) dbg(4, "ch1: level=%.2f", ch->level);
	return ch->level;
}


static double
am__mixer_pan_to_display(double internal)
{
	//translate from -1.0 - +1.0 to -10.0 - +10.0.

	if(internal < -1.0 || internal > 1.0) gwarn("ayyi pan value out of range: %.2f\n", internal);
	return internal * 10.0;
}


double
am_channel__pan_value(Channel* channel, bool* has_pan)
{
	AyyiChannel* ch = ayyi_mixer__channel_at_quiet(channel->core_index);
	g_return_val_if_fail(ch, 0.0);
	static int ddd = 0;

	if(ch->has_pan){
		if(has_pan) *has_pan = ch->has_pan;

		if((ch->shm_idx == 1) && !(ddd++%10)) dbg(4, "ch1: pan=%.2f", ch->pan);
		return am__mixer_pan_to_display(ch->pan);
	}
	return 0.0;
}


int
am_channel__get_input_idx(Channel* channel)
{
	AyyiTrack* trk = ayyi_song__audio_track_at(channel->core_index); //strictly, core_index is a mixer_track index.
	return trk ? trk->input_idx : -1;
}


void
am_channel__get_input_string(Channel* channel, char* name)
{
	//returns the "short" input connection name.

	AyyiTrack* trk_shared = ayyi_song__audio_track_at(channel->core_index); //strictly, core_index is a mixer_track index.
	g_return_if_fail(trk_shared);
	name[0] = '\0';
	if(trk_shared->input_idx < 1) return; //channel has no input connection.

	AyyiConnection* connection = ayyi_song__connection_at(trk_shared->input_idx);
	g_return_if_fail(connection);
	dbg(2, "input=%i name=%s", trk_shared->input_idx, connection->name);
	if(connection->io == 1) gwarn("output is an input!");

	char (*s)[64] = (char (*)[64])g_new0(char, 4 * 64);
	if(ayyi_connection__parse_string(connection, s)){
		strcpy(name, s[0]);
	}
	else gwarn("failed to parse connection string.");

	g_free(s);
}


void
am_channel__get_output_string(Channel* channel, char* output)
{
	//returns the "short" output connection name.
	//-we are returning what the output is connected to, so we are returning an _input_

	output[0] = '\0';

	AyyiTrack* ayyi_trk = ayyi_song__audio_track_at(channel->core_index); //strictly, core_index is a mixer_track index.
	g_return_if_fail(ayyi_trk);

	AyyiConnection* connection = ayyi_track__get_output(ayyi_trk);
	if(!connection || !connection->shm_idx) return; //channel has no output connection.

	//this is the connection we are connecting _to_, so it should be an input.
	if(connection->io == 0) gwarn("input is an output!");

	char (*s)[64] = (char (*)[64])g_new0(char, 4 * 64);
	if(ayyi_connection__parse_string(connection, s)){
		strcpy(output, s[0]);
	}
	else gwarn("failed to parse connection string.");

	g_free(s);
}


double
am_channel__get_aux_level(Channel* ch, int aux_num)
{
	//return the level in decibels

	AyyiAux* aux = ayyi_mixer__aux_get_(ch->core_index, aux_num);
	//if(aux) dbg(0, "level=%.2fdB(%.2f)", am_gain2db(aux->level), aux->level);
	return aux ? am_gain2db(aux->level) : 0.0;
}


void
am_channel__set_aux_level(Channel* ch, int aux_num, float value_db, AyyiHandler callback, gpointer user_data)
{
	uint32_t object_idx = ch->core_index | (aux_num << 16);

	AyyiAction* a = ayyi_action_new("set aux level (channel %i, aux %i) %.2fdB", ch->core_index, aux_num, value_db);
	a->callback   = ayyi_handler_simple;
	a->app_data   = ayyi_handler_data_new(callback, user_data);

	ayyi_set_float(a, AYYI_OBJECT_AUX, object_idx, AYYI_LEVEL, am_db2gain(value_db));
}


void
am_channel__set_aux_output(Channel* ch, int aux_num, int connection, AyyiHandler callback, gpointer user_data)
{
	// this will turn on the connection with the given shared->connection_index.
	// - if connection is -1, a new bus will be created.

	// the returned AyyiIdent is for the Channel.

	void am_channel__set_aux_output_done(AyyiAction* action)
	{
		PF;
		if(!action->ret.error){
			log_print(0, action->i_val ? "Aux bus set" : "Aux send disabled");
		}

		uint32_t ch_i = action->obj_idx.idx1 & 0xffff;
		int aux_num = action->obj_idx.idx1 >> 16;

		// check that ayyi_aux output was correctly set.
		AyyiAux* aux = ayyi_mixer__aux_get_(ch_i, aux_num);
		if(!aux){
			if(action->i_val) gwarn("couldnt get AyyiAux for channel=%i aux_num=%i tot_auxs=%i", action->obj_idx.idx1, aux_num, ayyi_mixer__aux_count(action->obj_idx.idx1));
		}
		if(aux) dbg(0, "aux_num=%i bus_num=%i gain=%.4f", aux_num, aux->bus_num, aux->level);

#if 0
		Channel* channel = (Channel*)am_array_at(song->channels, action->obj_idx.idx1);
//#warning emitting signal in non-standard place
		am_channels__emit("change", NULL, channel); //FIXME
#endif

		HandlerData* d = action->app_data;
		if(d){
			call(d->callback, (AyyiIdent){action->obj.type, action->obj_idx.idx1}, &action->ret.error, d->user_data);
		}
	}

	AyyiAction* a   = ayyi_action_new("set aux output (track %i, aux %i)", ch->core_index, aux_num);
	a->obj.type     = AYYI_OBJECT_AUX;
	a->prop         = AYYI_OUTPUT;
	a->obj_idx.idx1 = ch->core_index | (aux_num << 16);
	a->i_val        = connection;
	a->callback     = am_channel__set_aux_output_done;
	a->app_data     = ayyi_handler_data_new(callback, user_data);
	action_execute(a);
}


void
am_channel__set_enable_pan(Channel* ch, gboolean enable, AyyiHandler callback, gpointer user_data)
{
	g_return_if_fail(ch);

	AyyiAction* a   = ayyi_action_new("enable pan (channel %i) %i", ch->core_index, enable);
	a->obj.type     = AYYI_OBJECT_CHAN;
	a->obj_idx.idx1 = ch->core_index;
	a->ret.idx      = ch->core_index;
	a->prop         = AYYI_PAN_ENABLE;
	a->i_val        = enable;
	a->callback     = ayyi_handler_simple;
	a->app_data     = ayyi_handler_data_new(callback, user_data);

	action_execute(a);
}


AyyiConnection*
am_channel__get_connected_to (Channel* ch)
{
	//returns only the first routing connection
	//-api is provisional and is likely to change!

	AyyiTrack* ayyi_track = ayyi_song__audio_track_at(ch->core_index);
	g_return_val_if_fail(ayyi_track, NULL);

	AyyiList* routing = ayyi_list__first(ayyi_track->output_routing);

	if(routing) return ayyi_song__connection_at(routing->id);

	return NULL;
}


static GtkTreeModel*
am_channel_create_plugin_model()
{
	//TODO this model can probably be shared by all inserts on all channels..

	GtkTreeIter iter;

	GtkWidget* cellview = gtk_cell_view_new ();

	GtkTreeStore* store = gtk_tree_store_new (3, GDK_TYPE_PIXBUF, G_TYPE_STRING, G_TYPE_BOOLEAN);

	gtk_tree_store_append (store, &iter, NULL);
	gtk_tree_store_set (store, &iter,
                        0, blank_icon,
                        1, NO_PLUGIN,
                        2, FALSE,
                        -1);
	gtk_tree_store_append (store, &iter, NULL);
	gtk_tree_store_set (store, &iter,
                        0, blank_icon,
                        1, "Bypass",
                        2, FALSE,
                        -1);
	gtk_tree_store_append (store, &iter, NULL);
	gtk_tree_store_set (store, &iter,
                        0, blank_icon,
                        1, "Edit",
                        2, FALSE,
                        -1);
	gtk_tree_store_append (store, &iter, NULL);
	gtk_tree_store_set (store, &iter,
                        1, "Separator",
                        2, FALSE,
                        -1);

	GList* l = song->plugins;
	for(;l;l=l->next){
		AMPlugin* plug = l->data;

		gtk_tree_store_append (store, &iter, NULL);
		gtk_tree_store_set (store, &iter,
                        0, am_plugin__get_icon(plug),
                        1, plug->name,
                        2, FALSE,
                        -1);

		AyyiPlugin* plugin_shared = ayyi_plugin_at(plug->shm_num);
		if(plugin_shared){
			dbg (1, "category=%s", plugin_shared->category); //category = VST, etc
		}
	}

	gtk_widget_destroy (cellview);

	return GTK_TREE_MODEL (store);
}


bool
am_channel__is_type(Channel* ch, gpointer _type)
{
	ChannelType type = (ChannelType)_type;
	return ch->type == type;
}


