/*
  This file is part of the Ayyi Project. http://ayyi.org
  copyright (C) 2004-2012 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 3
  as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#define __am_private__
#include <model/common.h>
#include <ayyi/ayyi_list.h>
#include <ayyi/ayyi_shm.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_mixer.h>
#include <ayyi/ayyi_msg.h>
#include <ayyi/ayyi_action.h>
#include <ayyi/ayyi_dbus.h>
#include "ayyi/ayyi_dbus_proxy.h"

#include "model/am_plugin.h"
#include "model/am_track_peer.h"
#include "model/am_song.h"
#include "model/am_client.h"


AyyiClient*
am__init(AyyiService** services)
{
	// if services is null, the default Ardourd service will be used.

	static gboolean done = false;
	g_return_val_if_fail(!done, NULL);

	AyyiClient* a = ayyi_client_init();

	ayyi_client_add_services(services);

	ayyi_log_init();
	am_plugin__init();
	am_tracks_init();
	am_song__new();

	am_song__add_handlers();

	done = true;

	return a;
}


bool
am__connect_services(AyyiService** services, AyyiHandler2 on_connect, gpointer _user_data)
{
	// convenience function that will establish both messaging, and shm. The shm callback is only called once _all_ segments are attached.
	// -is the same as calling am__connect_messaging and am__get_shm.

	// the dbus connection is synchronous (fn will return imediately on failure), but shm connection is async.

	g_return_val_if_fail(services && services[0], false);

	typedef struct
	{
		AyyiHandler2  on_all_shm;
		gpointer      user_data;
		int           timeout;
		AyyiService** services;
		int           s;
		void          (*on_messaging)(GError*, gpointer);
		GError*       error;  //currently only the first error is forwarded.
	} C;
	C* c = g_new0(C, 1);
	c->on_all_shm = on_connect;
	c->user_data = _user_data;
	c->services = services;
	c->s = -1;

	void clear_timeout(C* c)
	{
		if(c->timeout) g_source_remove (c->timeout);
		c->timeout = 0;
	}

	gboolean am_connect_timeout(gpointer _c)
	{
		C* c = _c;
		call(c->on_all_shm, c->error, c->user_data);
		//TODO free c - check is always safe
		return TIMER_STOP;
	}
	c->timeout = g_timeout_add(4000, am_connect_timeout, c);

	void am_connect_all__on_shm(AyyiShmCallbackData* d)
	{
		C* c = d->user_data;

		if(!d->shm_seg){
			//its too late to set the GError here. Error is indicated by NULL segment.
			c->on_all_shm(c->error, c->user_data);
			//need to free c ?
			return;
		}

		bool
		service_shm_setup_is_complete(AyyiService* service)
		{
			dbg(1, "shm.song=%p ((AyyiSongService*)ayyi.service)->amixer=%p", ((AyyiSongService*)ayyi.service)->song, ((AyyiSongService*)ayyi.service)->amixer);

			if(service == known_services[AYYI_SERVICE_MIXER]){
				GList* l = service->segs;
				for(;l;l=l->next){
					shm_seg* seg = l->data;
					if(!seg->attached || seg->invalid){ dbg(2, "not complete."); return false; }
				}
				return g_list_length(service->segs) >= G_N_ELEMENTS(((AyyiMixerService*)service)->segments);
			}
			return ((((AyyiSongService*)service)->song) && (((AyyiSongService*)service)->amixer));
		}

		if(service_shm_setup_is_complete(d->service)){
			clear_timeout(c);
			//am__register_signals(ayyi.service);
			int i; for(i=0;c->services[i];i++){
				if(c->services[i] == known_services[AYYI_SERVICE_AYYID1]){ //FIXME create a service_load fn?
					am_song__load();
					break;
				}
			}
			c->on_all_shm(c->error, c->user_data);
			g_free(c);
		}
	}

	void am__next_service(C* c)
	{
GError** error = NULL;
		dbg(1, "s=%i", c->s + 1);
		if(c->services[++c->s]) am__connect_messaging(c->services[c->s], c->on_messaging, c, error);
		else dbg(1, "no more services");
	}

	void am__on_messaging(GError* error, gpointer user_data)
	{
		dbg(0, "connected! TODO handle failure");
		C* c = user_data;
		if(error){
			gwarn("error");
			c->error = g_error_copy(error);
		}else{
			AyyiService* service = c->services[c->s];
			//g_return_if_fail(service->priv->proxy);
			am__get_shm(service, am_connect_all__on_shm, c);
		}
		am__next_service(c);
	}
	c->on_messaging = am__on_messaging;

	gboolean connected = false;
#if 0
	int s; for(s=0;services[s];s++){
		AyyiService* service = services[s];

		if ((connected = am__connect_messaging(service, on_messaging, c, error))){
			am__get_shm(service, am_connect_all__on_shm, c);
		}
		else return false;
	}
#else
	am__next_service(c);
#endif
	return connected;
}


void
am__connect_all(AyyiHandler2 on_connect, gpointer _user_data)
{
	// convenience function that will establish both messaging, and shm for the Ardourd song and mixer services.
	// The shm callback is only called once _all_ segments are attached.
	// -is the same as calling am__connect_messaging and am__get_shm.

	// the dbus connection is synchronous (fn will return imediately on failure), but shm connection is async.

	// TODO instead, make this a wrapper for am__connect_services()

#if 0
	typedef struct
	{
		void     (*on_all_shm)(gpointer);
		gpointer user_data;
		int      timeout;
	} C;
	C* c = g_new0(C, 1);
	c->on_all_shm = on_connect;
	c->user_data = _user_data;

	void clear_timeout(C* c)
	{
		if(c->timeout) g_source_remove (c->timeout);
		c->timeout = 0;
	}

	gboolean am_connect_timeout(gpointer _c)
	{
		C* c = _c;
		call(c->on_all_shm, c->user_data);
		//TODO free c - is it always safe?
		return TIMER_STOP;
	}
	c->timeout = g_timeout_add(4000, am_connect_timeout, c);

	void am_connect_all__on_shm(AyyiShmCallbackData* d)
	{
		C* c = d->user_data;

		if(!d->shm_seg){
			//its too late to set the GError here. Error is indicated by NULL segment.
			c->on_all_shm(c->user_data);
			return;
		}

		bool
		service_shm_setup_is_complete(AyyiService* service)
		{
			dbg(1, "shm.song=%p ((AyyiSongService*)ayyi.service)->amixer=%p", ((AyyiSongService*)ayyi.service)->song, ((AyyiSongService*)ayyi.service)->amixer);

			return ((((AyyiSongService*)ayyi.service)->song) && (((AyyiSongService*)ayyi.service)->amixer));
		}

		if(service_shm_setup_is_complete(d->service)){
			clear_timeout(c);
			//am__register_signals(ayyi.service);
			am_song__load();
			c->on_all_shm(c->user_data);
		}
	}

	gboolean connected = false;

	if ((connected = am__connect_messaging(ayyi.service, NULL, NULL, error))){

		int type; for(type=1;type<=2;type++) ayyi_shm_seg_new(ayyi.service, type);

		am__get_shm(ayyi.service, am_connect_all__on_shm, c);
	}
	return connected;
#else
	AyyiService* services[] = {
		(AyyiService*)&ardourd_song,
		(AyyiService*)&ardourd_mixer,
		NULL,
	};

	typedef struct
	{
		AyyiHandler2 user_callback;
		gpointer     user_data;
		int          timeout;                         // TODO test
	} C;
	C* c = g_new0(C, 1);
	c->user_callback = on_connect;
	c->user_data = _user_data;

	void am__connect_connected(GError* error, gpointer _c)
	{
		C* c = _c;
		call(c->user_callback, error, c->user_data);
		g_free(c);
	}

	am__connect_services((AyyiService**)&services, am__connect_connected, c);
#endif
}


gboolean
am__connect_messaging(AyyiService* service, AyyiHandler2 callback, gpointer user_data, GError** error)
{
	//currently this should not to be used for connecting to plugins because of the signal registration.

	static bool signal_reg_done = false;

	HandlerData* c = g_new0(HandlerData, 1);
	c->callback = (AyyiHandler)callback;
	c->user_data = user_data;

	void am__connect_messaging__on_connected(GError* error, gpointer user_data)
	{
		PF;
		HandlerData* c = user_data;
		call((AyyiHandler2)c->callback, error, c->user_data);
		g_free(c);
	}

	if(ayyi_client_connect(service, am__connect_messaging__on_connected, c, error)){
		if(!signal_reg_done) ayyi_client_set_signal_reg(NEW | DELETED | CHANGED | TRANSPORT | LOCATORS | PROPERTY | PROGRESS);
		signal_reg_done = true;
		return true;
	}
	return false;
}


gboolean
am__get_shm(AyyiService* service, void (*on_shm)(AyyiShmCallbackData*), gpointer user_data)
{
	return ayyi_client__dbus_get_shm(service, on_shm, user_data);
}


gboolean
am__get_shm_(AyyiService** services, void (*on_shm)(AyyiShmCallbackData*), gpointer user_data)
{
	g_return_val_if_fail(services, false);

	int i; for(i=0;services[i];i++){
		if(!ayyi_client__dbus_get_shm(services[i], on_shm, user_data)) return false;
	}
	return true;
}


