#ifndef __pool_item_h__
#define __pool_item_h__

#include "model/model_types.h"
#include <model/am_peak.h>

struct _pool_item
{
	int                pod_index[2];    // -1 if not set.
	guint64            source_id[2];

	char               filename[256];   // this should be the exact name in the songcore.
	char               dirname [256];   // either relative or full path?
	char               leafname[256];   // the filename without directory info.
	int                state;
	gboolean           valid;           // true by default. False if the file is innaccessible. //DEPRECATED! - use state instead.

	gint64             samplecount;     // number of samples in the file.
	int                sample_rate;
	int                channels;
	int                bytes_per_sample;

	Waveform*          waveform;

	//int                peak_byte_depth; // 2 bytes by default, 4 bytes for float peaks.
	gboolean           peakfile_valid;

	void*              gui_ref;         // eg GtkTreeRowReference*

	float              max_db;

	void*              user_data;
};


enum
{
  POOL_ITEM_NEW = 0,
  POOL_ITEM_PENDING,   //waiting for remote confirmation
  POOL_ITEM_BAD,
  POOL_ITEM_INACCESSIBLE,
  POOL_ITEM_OK,
  POOL_ITEM_STATE_MAX
};


PoolItem*            pool_item_new                 ();
void                 pool_item_free                (PoolItem*);
PoolItem*            pool_item_new_from_ayyi       (AyyiFilesource*);
#ifdef __am_private__
void                 pool_item_set_from_ayyi       (PoolItem*, AyyiFilesource*);
#endif
bool                 pool_item_set_file            (PoolItem*, const char* fname, bool fast);
void                 pool_item_add_channel         (PoolItem*, AyyiFilesource*);
void                 pool_item_set_paths           (PoolItem*, char* path);
gboolean             pool_item_set_file_info       (PoolItem*);
void                 pool_item_sync                (PoolItem*);
gboolean             pool_item_file_exists         (PoolItem*);
GList*               pool_item_get_regions         (PoolItem*);
int                  pool_item_count_regions       (PoolItem*);
gboolean             pool_item_get_default_region  (PoolItem*, uint32_t* region_idx);
gboolean             pool_item_is_raw              (PoolItem*);
int                  pool_item_get_width           (PoolItem*);
char*                pool_item_get_full_path       (PoolItem*, int ch_num);
void                 pool_item_get_display_name    (PoolItem*, char* name);
gboolean             pool_item_in_core             (PoolItem*);
void                 pool_item_load_peak           (PoolItem*, int ch_num);
void                 pool_item_print               (PoolItem*);
const char*          pool_item_print_state         (PoolItem*);

bool                 pool_item_is_nonempty         (PoolItem*, gpointer);
bool                 pool_item_is_new              (PoolItem*, gpointer);

#endif //__pool_item_h__
