#ifndef __model_types_h__
#define __model_types_h__

#define AM_MAX_TRK 48
#define AM_MAX_CHS 16
#define MAX_PART_LENGTH_BEATS 2000

#define TIMER_CONTINUE TRUE
#define TIMER_STOP FALSE
#define IDLE_STOP FALSE
#define ZERO_TERMINATED FALSE

struct _gsong_pos
{
  int                beat;
  unsigned short int sub;           //subbeat (4 per beat) - note: not the same as core sub.
  unsigned short int tick;          //960 per sub, as displayed. Not the same as core mu.
};


#ifndef __GTK_H__
typedef void GtkStockItem;
#endif

struct s_key{
	int           code;
	int           mask;
};

struct _accel{                      //our version of GtkActionEntry - has two keys instead of one.
	char          name[16];
	struct s_key  key[2];
	gpointer      callback;
	GtkStockItem* stock_item;
	gpointer      user_data;
};

typedef enum {
  TRK_TYPE_NULL = 0,
  TRK_TYPE_AUDIO,
  TRK_TYPE_MIDI,
  TRK_TYPE_BUS,
} AMTrackType;

typedef enum {
  Q_16,
  Q_8,
  Q_BEAT,
  Q_BAR,
  Q_EXPONENTIAL,
  Q_MAX
} QType;

typedef enum{
  SNAP_OFF = 0,
  SNAP_BAR,
  SNAP_BEAT,
  SNAP_16,
  SNAP_Q,
  MAX_SNAP,
} AMSnapType;

typedef enum
{
  CHANGE_POS_X = 1,
  CHANGE_LEN   = 1 << 1,
  CHANGE_INSET = 1 << 2, //front trim
  CHANGE_WIDTH = 1 << 3,
  CHANGE_HEIGHT = 1 << 4,
  CHANGE_OUTPUT = 1 << 5,
  CHANGE_FADES = 1 << 6, //part fade in and out.
  CHANGE_NAME = 1 << 7,
  CHANGE_TRACK = 256,    //when applied to tracks, it refers to the track assignment for a widget.
  CHANGE_ZOOM_H = 512,
  CHANGE_COLOUR = 1024,
  CHANGE_ZOOM_V = 2048,
  CHANGE_ARMED = 2048 << 1,
  CHANGE_MUTE = 2048 << 2,
  CHANGE_SOLO = 4096 << 2,
  CHANGE_SELECTION = 4096 << 3,
  CHANGE_BACKGROUND_COLOUR = 4096 << 4,
  CHANGE_IDX = 4096 << 5,
  CHANGE_ALL = 0xffff,
} AMChangeType;


#ifndef bool
typedef gboolean bool;
#endif
typedef struct { int x; int y; } Pti;
typedef struct _ptf { float x; float y; } Ptf;
typedef struct _ptd { double x; double y; } Ptd;
typedef struct _drect { double x1, y1, x2, y2; } DRect;
typedef enum { MOVETO, MOVETO_OPEN, CURVETO, LINETO, BP_END } Pathcode;
typedef struct _bpath BPath;
struct _bpath { Pathcode code; double x1; double y1; double x2; double y2; double x3; double y3; };
typedef struct _vpath { Pathcode code; double x; double y; } VPath;
typedef struct { double r, g, b; } DColour;

#ifndef __ayyi_typedefs_h__
#define __ayyi_typedefs_h__
typedef struct _shm_seg            shm_seg;
typedef struct _ayyi_action        AyyiAction;
typedef struct _route_shared       AyyiTrack;
typedef struct _region_base_shared AyyiRegionBase;
typedef struct _region_shared      AyyiRegion;
typedef struct _region_shared      AyyiAudioRegion;
typedef struct _midi_region_shared AyyiMidiRegion;
typedef struct _song_pos           SongPos;
typedef struct _filesource_shared  filesource_shared;
typedef struct _ayyi_obj_idx       AyyiObjIdx;
typedef struct _ayyi_shm_seg       AyyiShmSeg;
typedef struct _ayyi_connection    AyyiConnection;
typedef struct block               block;
#endif
typedef struct _shm_seg_mixer      ShmSegMixer;
typedef struct _am_track           AMTrack;
typedef struct _channel            Channel;
typedef struct _am_part            AMPart;
typedef struct _part_midi          MidiPart;
typedef int                        TrackNum;
typedef struct _gsong_pos          GPos;
typedef struct _pool_item          PoolItem;
typedef struct _palette            Palette;
typedef struct _q_map              QMap;
typedef struct _curve              Curve;
typedef struct _observer           Observer;
typedef struct _plug               AMPlugin;
typedef struct _action             Action;
typedef struct _am_iter            AMIter;

typedef struct _AMArray            AMArray;
typedef struct _AMList             AMList;
typedef struct _AMTrackList        AMTrackList;
typedef struct _FilterIterator     FilterIterator;
typedef struct _IteratorFilterClass IteratorFilterClass;
typedef struct _IteratorFilterPrivate IteratorFilterPrivate;
typedef gboolean (*Filter) (void* item, void* user_data);

struct _am_iter
{
	int i;
	void* ptr;
	GList* list;
	GPtrArray* array;
	//void* (*next)(struct _am_iter*);
};

typedef struct
{
	MidiNote* orig;
	MidiNote  transient;
} TransitionalNote;

typedef struct _note_selection_list_item
{
	MidiPart* part;
	GList*    notes; // list of TransitionalNote*
} NoteSelectionListItem;

struct _q_map{
	char      name[64];
	//GPos      length;
	int       size;                 //the index of the last valid entry in the pos[] array.
	GPos      pos[33];              //the last entry in pos (pos[size]) should be equal to length.
};

#endif //__model_types_h__
