/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2009 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include <model/common.h>
#include <model/am_part.h>
#include <model/note_selection.h>


GList*
note_selection_list__new(GList** selection, AMPart* gpart)
{
	//copy the selected notes in the current Part to a new list of NoteSelectionListItem*.

	//currently the list only contains an entry for *1* Part.

	//lifetime: created lazily in _motion or _finish callbacks.
	//          should not be cleared at the end of each op so that in a subsequent
	//          op, the observer knows to clear the previous selection.

	if(*selection) gwarn("note_selection not previously reset.");

	NoteSelectionListItem* item = g_new0(NoteSelectionListItem, 1);

	GList* copy = NULL;
	GList* l = ((MidiPart*)gpart)->note_selection;
	for(;l;l=l->next){
		MidiNote* src = l->data;
		TransitionalNote* t = transitional_note__new(src);
		copy = g_list_append(copy, t);
	}
	dbg(2, "len=%i", g_list_length(copy));

	item->part = (MidiPart*)gpart;
	item->notes = copy;
	GList* list = g_list_append(NULL, item);

	return *selection = list;
}


void
note_selection_list__free(GList** selection)
{
	GList* l = *selection;
	for(;l;l=l->next){
		NoteSelectionListItem* item = l->data;
		GList* k = item->notes;
		for(;k;k=k->next){
			g_free((TransitionalNote*)k->data);
		}
		g_list_free(l->data);
	}
	list_clear(*selection);
}


void
note_selection_list__reset(GList* selection, AMPart* part)
{
	// removes the note list for each part but leaves the part item so the observer know which Parts to update.

	am_part__set_note_selection(part, NULL);
	GList* l = selection;
	for(;l;l=l->next){
		NoteSelectionListItem* item = l->data;
		GList* k = item->notes;
		for(;k;k=k->next){
			g_free((TransitionalNote*)k->data);
		}
		list_clear(item->notes);
	}
	dbg(0, "n_parts=%i", g_list_length(selection));
}


void
note_selection_list__print(GList** selection)
{
	UNDERLINE;
	if(!*selection) dbg(0, "empty.");

	GList* l = *selection;
	for(;l;l=l->next){
		NoteSelectionListItem* item = l->data;
		AMPart* part = (AMPart*)item->part;
		dbg(0, "part '%s':", part->name);
		GList* k = item->notes;
		if(!k) dbg(0, "  no notes selected.");
		for(;k;k=k->next){
			TransitionalNote* n = k->data;
			dbg(0, "  note=%i vel=%i", n->orig->note, n->orig->velocity);
		}
	}
	UNDERLINE;
}


/*static*/TransitionalNote*
transitional_note__new(MidiNote* note)
{
	TransitionalNote* t = g_new0(TransitionalNote, 1);
	t->orig = note;
	t->transient = *note;
	return t;
}


