typedef struct {
	const char* module_name;
	gpointer    user_data;
} TimeoutData;

#define START_MODULE \
	printf("starting module: %s%s%s ...\n", green, __func__, white); \
	g_return_if_fail(on_finish); \
	static gpointer __user_data; __user_data = _user_data; \
	static AyyiFinish __finish = NULL; \
	__finish = on_finish; \
	static int timer; \
	gboolean module_timeout(gpointer _d) \
	{ \
		printf("MODULE TIMEOUT! %s\n", ((TimeoutData*)_d)->module_name); \
		FAIL_TEST_TIMER("MODULE TIMEOUT!\n"); \
		g_free(_d); \
		return TIMER_STOP; \
	} \
	TimeoutData* __d = g_new0(TimeoutData, 1); /* TODO is not always freed. perhaps have a standard ClosureData struct for the whole module */ \
	__d->module_name = __func__; \
	__d->user_data = _user_data; \
	timer = g_timeout_add(40000, module_timeout, __d);

#define FINISH_MODULE2 \
	{ g_source_remove(timer); \
	if(d){ \
		if(d->on_finish) d->on_finish(d->user_data); \
		g_free(d); \
	}}

#define FINISH_MODULE(OBJ) \
	{ g_source_remove(timer); \
	__finish(OBJ, __user_data); }

#define FAIL_MODULE_IF_ERROR \
	if(error && *error){ g_source_remove(timer); FAIL_TEST("> %s", (*error)->message); }
/*
 *  Modules are called by test functions.
 *  -they can do anything they like as long as they call the callback when finished.
 *  -TODO define exactly what happens on failure (...currently we seem to just call FAIL_TEST directly ?).
 */

void
module__save_song(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	void module__on_save(AyyiIdent id, GError** error, gpointer user_data)
	{
		FINISH_MODULE(id);
	}

	am_song__save(module__on_save, NULL);
}


void
module__load_song_with_delay(AyyiFinish on_finish, gpointer _user_data, const char* _path)
{
	// does not save the current song.

	START_MODULE;

	dbg(0, "loading song: %s", _path);

	typedef struct
	{
		GCallback   on_unload;
		bool        unload_done;
		const char* path;
	} C;
	C* c = g_new0(C, 1);
	static C* cc; cc = c;
	c->path = g_strdup(_path);

	void module_on_unload(GObject* object, gpointer _c)
	{
		C* c = _c;
		assert((c == cc), "bad closure data");
		PF0;
		c->unload_done = true;
		g_signal_handlers_disconnect_by_func (song, c->on_unload, c);
	}

	c->on_unload = (GCallback)module_on_unload;

	void module_load__on_load__(AyyiIdent obj, GError** error, gpointer _c)
	{
		FAIL_MODULE_IF_ERROR;
		C* c = _c;
		assert((c == cc), "bad closure data");
		assert(song->loaded, "loaded flag not set");
		g_free((char*)c->path);
		g_free(c);
		FINISH_MODULE(AYYI_NULL_IDENT);
		PF;
	}

	am_song__connect("song-unload", G_CALLBACK(module_on_unload), c);
	am_song__load_new(c->path, module_load__on_load__, c);
}


void
module__save_and_load_song_with_delay(AyyiFinish on_finish, gpointer _user_data, const char* path)
{
	START_MODULE;
	/*
	static gboolean unloaded_done = false;

	void on_load(GObject* object, gpointer user_data)
	{
		gboolean load_pause_done(gpointer data)
		{
			PF;
			FINISH_MODULE(AYYI_NULL_IDENT);
			return STOP;
		}

		PF;
		if(!unloaded_done) gwarn("no unload signal received");
		assert(song->loaded, "");
		g_signal_handlers_disconnect_by_func (song, user_data, user_data);
		g_timeout_add(4000, load_pause_done, NULL);
	}
	*/
	void _load_finished(AyyiIdent obj, gpointer data)
	{
		PF;
		FINISH_MODULE(AYYI_NULL_IDENT);
	}

	void save_done(AyyiIdent obj, GError** error, gpointer data)
	{
		char* path = data;
		dbg(0, "path=%s", path);
		/*
		g_signal_connect(song, "song-load", G_CALLBACK(on_load), on_load);
		am_song__load_new(path, NULL, NULL);
		*/

		module__load_song_with_delay(_load_finished, NULL, path);
	}

	am_song__save(save_done, (void*)path);
}


#define remove_handler(OBJ, ID) {if(ID){ g_signal_handler_disconnect((gpointer)OBJ, ID); ID = 0; } else gwarn("disconnect: no handler id! %i", ID); }

void
module__add_audio_part(AyyiFinish on_finish, gpointer _user_data)
{
	//module to add an audio Part.
	//-the part counting test has been removed because it doesnt work with concurrent tests.

	START_MODULE;
	typedef struct {
		int handler_id;
		int check;
		bool got_signal;
	} C;
	C* c = g_new0(C, 1);
	c->check = 5;

	void
	module__add_audio_part_done(AyyiIdent obj, GError** error, gpointer user_data)
	{
		FAIL_MODULE_IF_ERROR;
		C* c = user_data;

		AyyiRegion* region = ayyi_song__audio_region_at(obj.idx);
		dbg(0, "idx=%i name='%s'", obj.idx, region ? region->name : NULL);
		assert(region, "region");

		AMPart* part = am_partmanager__get_by_id(region->id);
		assert(part, "cannot get AMPart");

		PoolItem* pool_item = part->pool_item;
		assert(pool_item, "PoolItem not set");

		assert(region->length > 1, "length not set");
		assert(region->length == pool_item->samplecount, "part length not correct: %i %Li", region->length, pool_item->samplecount);

		assert(c->got_signal, "signal not received");

		if(c->handler_id) remove_handler(song->parts, c->handler_id); //useless as should have been already done below
		g_free(c);
		FINISH_MODULE(obj);
	}

	void module__add_audio_part__on_have_pool_item(AyyiIdent obj, gpointer user_data)
	{
		void module__add_audio_part__on_have_track(AyyiIdent obj, gpointer _c)
		{
			C* c = _c;

			assert(ayyi_song__get_file_count(), "no files");

			//int n_tracks = model_collection_length(am_tracks);
			int n_tracks = ayyi_song__get_audio_track_count();
			int i = 0;
			int t = g_random_int_range(0, n_tracks); //FIXME not random
			AyyiTrack* track = NULL;
			while((track = ayyi_song__audio_track_next(track))){
				if(i >= t){
					if(strcmp(track->name, "master")) break;
				}
				i++;
			}
			if(!track) FAIL_TEST("no audio tracks!")
			dbg(2, "track=%i", track->shm_idx);

			char name[64];
			am_partmanager__make_name_unique(name, "New Part");

			void* ___user_data = NULL;
			FilterIterator* filter_iter = filter_iterator_new (0, NULL, NULL, am_pool, (Filter)pool_item_is_nonempty, ___user_data);
			PoolItem* pool_item = (PoolItem*)filter_iterator_next(filter_iter);
			filter_iterator_unref(filter_iter);
			assert(pool_item, "pool_item");

			void got_add_signal(GObject* _parts, AMPart* new_part, gpointer c)
			{
				assert((((C*)c)->check == 5), "closure check failed");
				((C*)c)->got_signal = true;
				remove_handler(song->parts, ((C*)c)->handler_id);
			}
			c->handler_id = g_signal_connect(song->parts, "add", G_CALLBACK(got_add_signal), c);
			dbg(2, "handler=%i", c->handler_id);

			uint32_t region_idx = REGION_FULL;
			uint64_t len = ayyi_samples2mu(pool_item->samplecount);
			AyyiSongPos start = {get_random_int(240), 0, 0};
			int n = 0;
			bool position_is_ok = false;
			while(true){
				//sometimes we fail to detect a part at a position
				//-is this because of rounding errors gpos -> cpos ?
				//- change to using just GPos? no, probably not accurate enough

				//*** in case of concurrent requests, a part can be created at a particular position between making the request and it being fulfilled.
				//TODO should be model create proxy objects for requests? ideally, yes.
				//     -we already have the song->parts_pending list but it is not integrated properly. move it into AM.List<AM.Part*>* song->parts ? <---- yes
				//     -can we make local objects proper first class objects ?
				//         -all access would have to be via AM and not Ayyi api (well, to include local-only parts anyway).
				//

				FilterIterator* filter_iter = filter_iterator_new (0, NULL, NULL, am_parts, (Filter)am_part__is_at_position, &start);
				AMPart* part = (AMPart*)filter_iterator_next(filter_iter);
				if(part){
					//there is a part here already
					start.beat = get_random_int(160);
					start.sub = get_random_int(4);
					dbg(0, "new start=%i", start.beat);
				}else{
					//there are no parts here.
					dbg(2, "no parts here");
					position_is_ok = true;
				}
				filter_iterator_unref(filter_iter);
				if(position_is_ok) break;
				n++;
				if(n > 511){
					FAIL_TEST("cannot find position");
					break;
				}
			}
			{
				GPos s;
				songpos_ayyi2gui(&s, &start);
				char bbst1[64]; pos2bbst(&s, bbst1); dbg(2, "going with: %s (%i) %i", bbst1, pos2samples(&s), ayyi_pos2samples(&start));
			}
			am_song__add_part(AYYI_AUDIO, region_idx, pool_item, len, track->shm_idx, &start, 0, (char*)name, 0, module__add_audio_part_done, c);
		}

		module__ensure_audio_track(module__add_audio_part__on_have_track, user_data);
	}
	module__ensure_pool_item(module__add_audio_part__on_have_pool_item, c);
}


void
module__add_n_audio_parts(AyyiFinish on_finish, int n, gpointer _user_data)
{
	START_MODULE;
	static int target_n_regions = 0; target_n_regions = n;

	struct _cb
	{
		void (*callback)(AyyiIdent, struct _cb*);
	};

	void _add_part_finished(AyyiIdent obj, struct _cb* data)
	{
		assert(data->callback, "");
		dbg(0, "%i > %i ?", ayyi_song__get_audio_region_count(), target_n_regions);
		if(ayyi_song__get_audio_region_count() >= target_n_regions) dbg(0, "target reached. Finishing...");
		if(ayyi_song__get_audio_region_count() >= target_n_regions){
			FINISH_MODULE(obj);
			return;
		}

		module__add_audio_part((gpointer)data->callback, data);
	}

	struct _cb* data = g_new0(struct _cb, 1);
	data->callback = _add_part_finished;

	module__add_audio_part((gpointer)_add_part_finished, data);
}


void
module__delete_audio_part(AyyiFinish on_finish, AMPart* part, gpointer _user_data)
{
	START_MODULE;
	//static bool got_signal; got_signal = false;
	typedef struct _closure {
		int   handler;
		void* callback;
	} Closure;
	Closure* c = g_new0(Closure, 1);

	static int n_parts; n_parts = model_collection_length(am_parts);
	g_return_if_fail(part);

	void module__delete_audio_part_done(AyyiIdent id, GError** error, gpointer user_data)
	{
		Closure* c = user_data;
		dbg(0, "n_parts=%i-->%i", n_parts, model_collection_length(am_parts));
		if(n_parts - 1 != model_collection_length(am_parts)) FAIL_TEST("wrong part count");
		if(c->handler) remove_handler(song->parts, c->handler);
		gulong n = g_signal_handler_find(song->parts, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, c->callback, NULL);
		dbg(0, "handler found: %lu (should be 0)", n);
		FINISH_MODULE(id);
	}

	void got_del_signal(GObject* _parts, AMPart* new_part, gpointer user_data)
	{
		PF;
		Closure* c = user_data;
		//got_signal = true;
		remove_handler(song->parts, c->handler);
		gulong n = g_signal_handler_find(song->parts, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, c->callback, NULL);
		dbg(0, "handler found: %lu (should be 0)", n);
	}
	c->handler = g_signal_connect(song->parts, "delete", G_CALLBACK(got_del_signal), c);
	c->callback = got_del_signal;
	gulong n = g_signal_handler_find(song->parts, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, got_del_signal, NULL);
	dbg(0, "handler found: %lu", n);

	am_song__remove_part(part, module__delete_audio_part_done, c);
}


void
module__delete_random_part(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	void module__delete_random_part_done(AyyiIdent id, GError** error, gpointer user_data)
	{
		FAIL_MODULE_IF_ERROR;
		FINISH_MODULE(id);
	}

	GRand* rand = g_rand_new();
	int r = g_rand_int_range(rand, 0, model_collection_length(am_parts) - 1);

	AMPart* part = g_list_nth_data(song->parts->list, r);
	assert(part, "part");

	am_song__remove_part(part, module__delete_random_part_done, NULL);
}


void
module__delete_all_parts(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	typedef struct {
		int handler_id;
		int n_deleted;
		int n_to_delete;
		bool finished;
	} C;
	C* c = g_new0(C, 1);

	void _finish(C* c)
	{
//its possible this handler might not be being removed properly.
dbg(0, "handler=%i removing...", c->handler_id);
		remove_handler(am_parts, c->handler_id);
		assert(!c->finished, "module already finished!");
		if(model_collection_length(am_parts)) am_partmanager__print_list();
		assert(model_collection_length(am_parts) == 0, "n_parts == 0");
		c->finished = true;
		g_free(c);
		FINISH_MODULE(AYYI_NULL_IDENT);
	}

	void _delete_all_audio_parts__on_delete(GObject* _song, AMPart* part, gpointer user_data)
	{
		C* c = user_data;
		c->n_deleted++;
		dbg(0, "%i parts deleted", c->n_deleted);
#ifdef DELETE_PARTS_BY_MODEL_LIST
		if(!model_collection_length(am_parts)){
#else
		if(!ayyi_song__get_audio_region_count() && !ayyi_song__get_midi_region_count()){
#endif

#if 0
			remove_handler(am_parts, c->handler_id);
			assert(!c->finished, "module already finished!");
			c->finished = true;
			g_free(c);
			FINISH_MODULE(AYYI_NULL_IDENT);
#endif
			_finish(c);
		}
	}

#ifdef DELETE_PARTS_BY_MODEL_LIST //this will _not_ delete regions that are not currently assigned to tracks.
	c->handler_id = g_signal_connect(am_parts, "delete", G_CALLBACK(_delete_all_audio_parts__on_delete), c);
//dbg(0, "handler=%i", c->handler_id);

	GList* l = song->parts->list;
	c->n_to_delete = g_list_length(l);
	for(;l;l=l->next){
		AMPart* part = l->data;
		dbg(0, "  removing: %i", part->ayyi->shm_idx);
		if(PART_IS_AUDIO(part))
			ayyi_song__delete_audio_region((AyyiAudioRegion*)part->ayyi);
		else
			ayyi_song__delete_midi_region((AyyiMidiRegion*)part->ayyi);
	}
	dbg(0, "parts found: %i", g_list_length(song->parts->list));
#else
	c->n_to_delete = 0;
	AyyiRegion* part = NULL;
	while((part = ayyi_song__audio_region_next(part))){
		if(!(part->flags & deleted)){
			c->n_to_delete++;
			ayyi_song__delete_audio_region(part);
		}
	}
	{
	AyyiMidiRegion* part = NULL;
	while((part = ayyi_song__midi_region_next(part))){
		if(!(part->flags & deleted)){
			c->n_to_delete++;
			ayyi_song__delete_midi_region((AyyiMidiRegion*)part);
		}
	}
	}
#endif

	gboolean container_is_empty(gpointer _c)
	{
		dbg(2, "n_regions=%i+%i", ayyi_song__get_audio_region_count(), ayyi_song__get_midi_region_count());
		return ayyi_container_is_empty(&((AyyiSongService*)ayyi.service)->song->audio_regions) && ayyi_container_is_empty(&((AyyiSongService*)ayyi.service)->song->midi_regions);
	}

	void wait_done(AyyiIdent id, GError** e, gpointer _c)
	{
		PF0;
		_finish((C*)_c);
	}

	//reset_timeout(30000);
	if(c->n_to_delete)
		am_wait_for(container_is_empty, wait_done, c);
	else
		_finish(c);
}


void
module__select_part(AyyiFinish on_finish, AMPart* part, gpointer _user_data)
{
	//if part is NULL, use first available part.

	START_MODULE;
	static int handler; handler = 0;
	//static bool got_signal; got_signal = false;

	assert(song->parts->list, "no parts");

	if(!part){
		part = song->parts->list->data;
	}

	void got_add_signal(GObject* _parts, AMPart* new_part, gpointer user_data)
	{
		//got_signal = true;
		remove_handler(song->parts, handler);
		FINISH_MODULE(AYYI_NULL_IDENT);
	}
	handler = g_signal_connect(song->parts, "selection-change", G_CALLBACK(got_add_signal), NULL);

	model_collection_selection_replace (am_parts, g_list_append(NULL, part), NULL);
}


void
module__split_part(AyyiFinish on_finish, AMPart* part, gpointer _user_data)
{
	//split a part in half.

	START_MODULE;

	assert(part, "part");
	dbg(0, "%s: len=%i", part->name, pos2samples(&part->length));
	assert(pos2samples(&part->length) > 1, "part too short to split");

	static int n_parts; n_parts = g_list_length(song->parts->list);

	void _module_split_part_finished(AyyiIdent obj, GError** error, gpointer data)
	{
		FAIL_MODULE_IF_ERROR;
		assert(g_list_length(song->parts->list) == n_parts + 1, "incorrect number of parts. %i != %i", g_list_length(song->parts->list), n_parts + 1);
		PF;
		FINISH_MODULE(obj);
	}

	GPos pos = part->length;
	pos_divide(&pos, 2);
	pos_add(&pos, &part->start);

	char start[32];
	char len[32];
	char end[32];
	pos2bbst(&part->start, start);
	pos2bbst0(&part->length, len);
	pos2bbst(&pos, end);
	dbg(0, "start=%s length=%s cut=%s", start, len, end);

	assert(pos_is_after(&pos, &part->start), "cut at start");

	AyyiSongPos p;
	songpos_gui2ayyi(&p, &pos);

	am_part__split(part, &p, _module_split_part_finished, NULL);
}


void
module__select_track(AyyiFinish on_finish, AMTrack* track, gpointer _user_data)
{
	//if part is NULL, use first available part.

	START_MODULE;
	static int handler; handler = 0;
	//static bool got_signal; got_signal = false;

	assert(am_track_list_at(song->tracks, 0), "no tracks");

	if(!track){
		track = am_track_list_at(song->tracks, 0);
	}

	void got_selection_signal(GObject* _parts, AMTrack* new_track, gpointer user_data)
	{
		//got_signal = true;
		remove_handler(song->tracks, handler);
		FINISH_MODULE(AYYI_NULL_IDENT);
	}
	handler = g_signal_connect(song->tracks, "selection-change", G_CALLBACK(got_selection_signal), NULL);

	model_collection_selection_replace (am_tracks, g_list_append(NULL, track), NULL);
}


void
module__add_audio_track(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;
	static bool got_signal = false;
	static int handler_id;

	void
	module_on_track_create(AyyiIdent id, GError** error, gpointer user_data)
	{
		FAIL_MODULE_IF_ERROR

		assert(got_signal, "got signal");
		remove_handler(song->channels, handler_id);

		AyyiIdx new_object_idx = id.idx;
		AyyiTrack* track = ayyi_song__audio_track_at(new_object_idx);
		dbg(0, "idx=%i name='%s'", new_object_idx, track ? track->name : NULL);

		assert(ayyi_verify_playlists(), "playlist verify");

		FINISH_MODULE(id)
	}

	void got_add_signal(GObject* _song, Channel* new_channel, gpointer user_data)
	{
		got_signal = true;
		remove_handler(song->channels, handler_id);
	}

	handler_id = g_signal_connect(song->channels, "add", G_CALLBACK(got_add_signal), NULL);

	char* name = g_strdup("New Track");
	am_song__add_track(TRK_TYPE_AUDIO, name, 1, module_on_track_create, _user_data);
}


void
module__delete_audio_track(AyyiFinish on_finish, gpointer _user_data, AMTrack* trk)
{
	START_MODULE;
	typedef struct {
		int      handler_id;
		bool     got_signal;
		gpointer user_data;
	} C;
	C* c = g_new0(C, 1);
	c->user_data = _user_data;

	void module__on_track_delete(AyyiIdent id, GError** error, gpointer user_data)
	{
		PF;
		C* c = user_data;
		FAIL_MODULE_IF_ERROR

		assert(c->got_signal, "got signal");
		remove_handler(song->channels, c->handler_id);
		g_free(c);

		assert(ayyi_verify_playlists(), "playlist verify");

		//iterating over the tracks will check that the ayyi route exists
		AMTrack* trk = NULL;
		AMIter i;
		model_collection_iter_init (am_tracks, &i);
		while((trk = model_collection_iter_next(am_tracks, &i))){
		}

		FINISH_MODULE(id);
	}

	void got_remove_signal(GObject* _song, Channel* new_channel, gpointer user_data)
	{
		PF;
		C* c = user_data;
		c->got_signal = true;
		remove_handler(song->channels, c->handler_id);
	}

	//this is an additional test, to make sure the mixer is properly updated.
	c->handler_id = g_signal_connect(song->channels, "delete", G_CALLBACK(got_remove_signal), c);

	am_song__remove_track(trk, module__on_track_delete, c);
}


void
module__delete_tracks_by_name(AyyiFinish on_finish, gpointer _user_data, char** track_names)
{
	START_MODULE;
	FINISH_MODULE(AYYI_NULL_IDENT);
}


void
module__delete_all_tracks(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	static int n_deleted; n_deleted = 0;
	static int n_to_delete; n_to_delete = 0;

	void module__on_delete(AyyiIdent id, GError** error, gpointer data)
	{
		FAIL_MODULE_IF_ERROR
		assert(ayyi_verify_playlists(), "playlist verify");
		n_deleted++;
		dbg(0, "%i of %i tracks deleted", n_deleted, n_to_delete);
		if(n_deleted == n_to_delete){
			assert(am_track_list_get_visible_count(song->tracks) == MIN_TRACKS, "tracks not cleared: %i", am_track_list_get_visible_count(song->tracks));
			assert(ayyi_song__get_playlist_count() == MIN_TRACKS, "playlists not cleared (%i != %i)", ayyi_song__get_playlist_count(), MIN_TRACKS);
			FINISH_MODULE(id);
		}
	}

	void module__delete_all_tracks__on_parts_deleted(AyyiIdent id, gpointer data)
	{
		assert(am_track_list_get_visible_count(song->tracks) == ayyi_song__get_playlist_count(), "playlist mismatch %i %i", am_track_list_get_visible_count(song->tracks), ayyi_song__get_playlist_count());
		assert(ayyi_verify_playlists(), "playlist verify");

		AMTrack* trk = NULL;
		AMIter i;
		model_collection_iter_init (am_tracks, &i);
		int n = 0;
		while((trk = model_collection_iter_next(am_tracks, &i))){
			n++;
			if(n < 2) continue; //leave one track remaining.
			n_to_delete++;
			dbg(0, "  removing: %i type=%i %s", trk->track_num, trk->type, trk->name);
			am_song__remove_track(trk, module__on_delete, data);
		}
		dbg(0, "tracks found: %i (%i)", n_to_delete, n);

		if(!n_to_delete) FINISH_MODULE(id);
	}

	assert(ayyi_verify_playlists(), "playlists not verified");

	//its not clear whether removing a tracks parts is neccesary.
	module__delete_all_parts(module__delete_all_tracks__on_parts_deleted, _user_data);
}


void
module__add_midi_track(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;
	assert(ayyi_verify_playlists(), "playlist verify");

	void test_add_midi_track_done(AyyiIdent id, GError** error, gpointer user_data)
	{
		FAIL_MODULE_IF_ERROR;
		assert(ayyi_song__get_midi_track_count(), "no midi tracks found");
		dbg(0, "n_midi_tracks=%i", ayyi_song__get_midi_track_count());
		AMTrack* track = am_track_list_find_by_ident(song->tracks, id);
		assert(track, "track not found");
		assert(track->type == TRK_TYPE_MIDI, "wrong track type")
		gboolean playlists_ok = ayyi_verify_playlists();
		if(!playlists_ok) ayyi_song__print_playlists();
		assert(playlists_ok, "playlist verify");
		FINISH_MODULE(id);
	}

	char unique[256];
	ayyi_track__make_name_unique(unique, "Test Midi Track");

	am_song__add_track(TRK_TYPE_MIDI, g_strdup(unique), 1, test_add_midi_track_done, _user_data);
}


void
module__add_midi_part(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	typedef struct {
		int handler_id;
		bool got_signal;
	} C;
	C* c = g_new0(C, 1);


	void module__add_midi_part__on_have_track(AyyiIdent obj, gpointer _c)
	{
		C* c = _c;

		AyyiMidiTrack* track = ayyi_song__midi_track_next (NULL);
		assert(track, "no tracks");

		void got_add_signal(GObject* _parts, AMPart* new_part, gpointer _c)
		{
			C* c = _c;
			c->got_signal = true;
			remove_handler(song->parts, c->handler_id);
			c->handler_id = 0;
		}
		c->handler_id = g_signal_connect(song->parts, "add", G_CALLBACK(got_add_signal), c);
		dbg(2, "handler=%i", c->handler_id);

		void module__add_midi_part_done (AyyiIdent obj, GError** error, gpointer _c)
		{
			PF;
			C* c = _c;
			assert(c->got_signal, "signal not received");

			if(c->handler_id) remove_handler(song->parts, c->handler_id);
			g_free(c);
			FINISH_MODULE(obj);
		}

		char* name = ayyi_song_container_next_name (&((AyyiSongService*)ayyi.service)->song->midi_regions, "Midi Part");
		GPos l = {4, 0, 0};
		uint64_t len = pos2mu (&l);
		AyyiSongPos pos = {1, 0, 0};

		am_song__add_part(AYYI_MIDI, -1, NULL, len, track->shm_idx, &pos, 0, name, 0, module__add_midi_part_done, c);
	}
	module__ensure_midi_track(module__add_midi_part__on_have_track, c);
}


void
module__delete_midi_part(AyyiFinish on_finish, AMPart* part, gpointer _user_data)
{
	START_MODULE;
	assert(part, "no part arg");

	void module__delete_midi_part_done(AyyiIdent obj, GError** error, gpointer data)
	{
		FAIL_MODULE_IF_ERROR;
		AMPart* part = data;
		FINISH_MODULE(part->ident);
	}
	am_song__remove_part(part, module__delete_midi_part_done, part);
}


void
module__add_midi_notes(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	void module__add_midi_notes_done(AyyiIdent id, GError** error, gpointer user)
	{
		PF;
		FAIL_MODULE_IF_ERROR;
		FINISH_MODULE(id);
	}

	AMPart* part = NULL;
	GList* l = song->parts->list;
	for(;l;l=l->next){
		if(PART_IS_MIDI((AMPart*)l->data)){
			part = l->data;
			break;
		}
	}
	assert(part, "part");

	GPos _start = {0, 1, 0};
	GPos _length = {0, 1, 0};
	MidiNote* note = ayyi_midi_note_new();
	note->note = 60;
	note->velocity = 100;
	note->start = pos2samples(&_start);
	note->length = (int)pos2samples(&_length);

	am_song__midi_note_add_async(part, note, module__add_midi_notes_done, NULL);
}


	typedef struct _module_data {
		AyyiFinish  on_finish;
		gpointer    user_data;
		const char* name;
	} ModuleData;
void
module__load_wav_fixture(AyyiFinish on_finish, gpointer _user_data, const char* _filename)
{
	START_MODULE;

	char* filename = NULL;
	filename = g_path_is_absolute(_filename)
		? ((char*)_filename)
		: g_build_filename(g_get_current_dir(), _filename, NULL);

	ModuleData* d = g_new(ModuleData, 1);
	d->on_finish = on_finish;
	d->user_data = _user_data;
	d->name = filename;

	void module__load_wav_fixture_done(AyyiIdent id, GError** error, gpointer user_data)
	{
		PF;
		FAIL_MODULE_IF_ERROR;
		ModuleData* d = user_data;

		assert(ayyi_song__have_file(d->name), "file not loaded");

		dbg(0, "n_files=%i", g_list_length(song->pool->list));
		assert(g_list_length(song->pool->list) > 0, "pool empty");

		FINISH_MODULE(id)
	}

	am_song__add_file(filename, module__load_wav_fixture_done, d);
}


void
module__delete_audio_file (AyyiFinish on_finish, gpointer _user_data, const char* filename)
{
	START_MODULE;
	assert(filename && strlen(filename), "filename not specified");
	dbg(0, "%s", filename);

	if(!g_list_length(song->pool->list)) FINISH_MODULE(AYYI_NULL_IDENT);

	typedef struct _data {
		const char* filename;
	} Closure;

	void module__delete_audio_file_done(AyyiIdent id, GError** error, gpointer data)
	{
		PF;
		FAIL_MODULE_IF_ERROR;

		AyyiFilesource* file = ayyi_filesource_get_by_name(((Closure*)data)->filename);
		assert(!file, "file not cleared");

		FINISH_MODULE(id);
	}

	struct _data* d = g_new0(Closure, 1);
	d->filename = filename;

	PoolItem* p = am_pool__get_item_by_orig_name (filename);
	dbg(0, "p=%p", p);
	assert(p, "cannot find pool_item");

	am_song__remove_file(p, module__delete_audio_file_done, d);
}


void
module__delete_all_audio_files (AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	dbg(0, "n_files=%i", model_collection_length(am_pool));
	if(!model_collection_length(am_pool)) FINISH_MODULE(AYYI_NULL_IDENT);

	void module__delete_all_audio_files_done(AyyiIdent id, GError** error, gpointer data)
	{
		PF;
		int n_files = g_list_length(song->pool->list);
		dbg(0, "n_files=%i", n_files);
		if(!n_files){
			FilterIterator* i = filter_iterator_new (0, NULL, NULL, am_parts, am_part__is_audio, NULL);
			AMPart* part = (AMPart*)filter_iterator_next(i);
			filter_iterator_unref(i);

			assert(!part, "n audio parts not zero");
			FINISH_MODULE(id)
		}
	}

	GList* l = song->pool->list;
	for(;l;l=l->next){ //warning! iterating over a list that is changing.
		am_song__remove_file((PoolItem*)l->data, module__delete_all_audio_files_done, NULL);
	}
}


void
module__delete_parts_by_audio_file (AyyiFinish on_finish, gpointer _user_data, PoolItem* pool_item)
{
	START_MODULE;

	void module__delete_parts_by_audio_file_done(AyyiIdent id, GError** error, gpointer data)
	{
		FAIL_MODULE_IF_ERROR;
		FINISH_MODULE(id)
	}

	am_song__remove_parts_by_file(pool_item, module__delete_parts_by_audio_file_done, NULL);
}


void
module__add_plugin(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	void module__add_plugin_done(AyyiIdent id, GError** error, gpointer data)
	{
		PF;
		//dbg(0, "error=%p", error ? *error : NULL);
		FAIL_MODULE_IF_ERROR;
		FINISH_MODULE(id)
	}

	//AMIter i;
	//channel_iter_init(&i);
	//Channel* channel = channel_next(&i);
	FilterIterator* i = filter_iterator_new (0, NULL, NULL, am_channels, (Filter)am_channel__is_type, (gpointer)CHAN_INPUT);
	Channel* channel = (Channel*)filter_iterator_next(i);
	filter_iterator_unref(i);

	AyyiIdx plugin_slot = 0;
	const char* plugin_name = vst_plugin_fixture;

	am_plugin__change(channel, plugin_slot, plugin_name, module__add_plugin_done, NULL);
}


void
module__ensure_pool_item(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;
	if(ayyi_song__get_file_count()){
		FINISH_MODULE(AYYI_NULL_IDENT)
		return;
	}

	dbg(0, "wav fixture being loading...");
	void on_wav_added(AyyiIdent obj, gpointer data)
	{
		PF;
		FINISH_MODULE(obj)
	}
	module__load_wav_fixture(on_wav_added, NULL, mono_wav_fixture);
}


void
module__ensure_audio_track(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	AMTrack* track = am_track_list_get_first_visible(song->tracks, TRK_TYPE_AUDIO);
	if(track){
		FINISH_MODULE(track->ident)
	}else{

		void module__ensure_audio_track_done(AyyiIdent obj, gpointer data)
		{
			FINISH_MODULE(obj)
		}

		module__add_audio_track(module__ensure_audio_track_done, NULL);
	}
}


void
module__ensure_mono_audio_track(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	FilterIterator* i = filter_iterator_new (0, NULL, NULL, am_tracks, (Filter)am_track__is_mono, NULL);
	AMTrack* track = (AMTrack*)filter_iterator_next(i);
	filter_iterator_unref(i);

	if(track){
		FINISH_MODULE(track->ident)
	}else{
		void module__ensure_mono_audio_track_done(AyyiIdent obj, gpointer data)
		{
			FINISH_MODULE(obj)
		}

		module__add_audio_track(module__ensure_mono_audio_track_done, NULL);
	}
}


void
module__ensure_pannable_track(AyyiFinish on_finish, gpointer _user_data)
{
	//currently only mono tracks are pannable.
	//the first mono track will be used, and a panner will be added to it if necessary.

	START_MODULE;

	FilterIterator* i = filter_iterator_new (0, NULL, NULL, am_tracks, (Filter)am_track__is_mono, NULL);
	AMTrack* track = (AMTrack*)filter_iterator_next(i);
	filter_iterator_unref(i);

	void module__ensure_mono_audio_track_done(AyyiIdent obj, gpointer data)
	{
		Channel* ch = channels__get_by_index(obj.idx);
		assert(ch, "channel not found");
		AyyiChannel* ac = ayyi_mixer__channel_at(obj.idx);
		assert(ac, "ayyi channel not found");

		void module__ensure_pannable_track_done(AyyiIdent obj, GError** error, gpointer data)
		{
			PF0;
			dbg(0, "idx=%i", obj.idx);
			FAIL_MODULE_IF_ERROR;
			FINISH_MODULE(obj)
		}

		dbg(0, "ch=%i has_pan=%i", obj.idx, ac->has_pan);
		if(ac->has_pan){
			FINISH_MODULE(obj);
		}else{
			/*  //can delete
			AMTrack* track = am_track_list_find_audio_by_idx(song->tracks, obj.idx);
			AyyiConnection* connection = NULL;
			while((connection = ayyi_connection__next_input(connection, 2))){
				break;
			}
			dbg(0, "connection=%s", connection->name);

			void done(AyyiIdent obj, GError** error, gpointer data) {}
			am_track__set_output(track, connection, done, NULL);
			*/

			am_channel__set_enable_pan(ch, true, module__ensure_pannable_track_done, NULL);
		}
	}

	if(track){
		module__ensure_mono_audio_track_done(track->ident, NULL);
	}else{
		module__add_audio_track(module__ensure_mono_audio_track_done, NULL);
	}
}


void
module__ensure_midi_track(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;

	AMTrack* midi_track = am_track_list_get_first_visible(song->tracks, TRK_TYPE_MIDI);
	if(midi_track){
		FINISH_MODULE(midi_track->ident)
	}else{

		void module__ensure_midi_track_done(AyyiIdent obj, gpointer data)
		{
			FINISH_MODULE(obj)
		}

		module__add_midi_track(module__ensure_midi_track_done, NULL);
	}
}


void
module__ensure_track_has_part(AyyiFinish on_finish, AMTrack* track, gpointer _user_data)
{
	START_MODULE;
	void module__ensure_track_has_part_done(gpointer data)
	{
		AMPart* part = data;
		FINISH_MODULE(part->ident);
	}

	GList* parts = am_partmanager__get_by_track(track);
	if(parts){
		AMPart* part = parts->data;
		g_list_free(parts);
		module__ensure_track_has_part_done(part);
	}else{
		void module__ensure_track_has_part__part_added(AyyiIdent id, GError** error, gpointer data)
		{
			AMPart* part = am_list_find_by_idx(song->parts, id.idx);
			assert(part, "part not found for idx=%i", id.idx);
			module__ensure_track_has_part_done(part);
		}
		PoolItem* pool_item = am_list_first (song->pool);
		assert(pool_item, "no pool_item");
		uint64_t len = ayyi_samples2mu(pool_item->samplecount);
		AyyiSongPos start = {get_random_int(40), 0, 0};
		am_song__add_part(AYYI_AUDIO, REGION_FULL, pool_item, len, track->ident.idx, &start, 0, track->name, 0, module__ensure_track_has_part__part_added, track);
	}
}


void
module__ensure_part(AyyiFinish on_finish, gpointer _user_data)
{
	START_MODULE;
	void module__ensure_part_done(gpointer data)
	{
		AMPart* part = data;
		FINISH_MODULE(part->ident);
	}

	AMPart* part = am_list_first(song->parts);
	if(part){
		module__ensure_part_done(part);
	}else{
		void module__ensure_part_have_poolitem(AyyiIdent id, gpointer data)
		{
			void module__ensure_part__part_added(AyyiIdent id, GError** error, gpointer data)
			{
				AMPart* part = am_list_find_by_idx(song->parts, id.idx);
				assert(part, "part not found for idx=%i", id.idx);
				module__ensure_part_done(part);
			}
			AMTrack* track = am_track_list_get_first_visible(song->tracks, TRK_TYPE_AUDIO);
			assert(track, "no tracks")
			PoolItem* pool_item = am_list_first (song->pool);
			assert(pool_item, "no pool_item");
			uint64_t len = ayyi_samples2mu(pool_item->samplecount);
			AyyiSongPos start = {get_random_int(40), 0, 0};
			am_song__add_part(AYYI_AUDIO, REGION_FULL, pool_item, len, track->ident.idx, &start, 0, track->name, 0, module__ensure_part__part_added, track);
		}

		module__ensure_pool_item(module__ensure_part_have_poolitem, NULL);
	}
}


void
module__ensure_aux(AyyiFinish on_finish, AMTrack* track, gpointer _user_data)
{
	//if track is NULL use first audio track.

	START_MODULE;

	void on_have_track(AyyiIdent t, gpointer user_data)
	{
		void on_have_aux(AyyiIdent trk, GError** error, gpointer user_data)
		{
			FAIL_MODULE_IF_ERROR;
			FINISH_MODULE(trk);
		}

		AMTrack* track = am_track_list_find_by_ident(song->tracks, t);
		assert(track, "track not found");
		Channel* ch = am_track__lookup_channel(track);

		AyyiChannel* chan = ayyi_mixer__channel_at_quiet(ch->core_index);
		AyyiAux* aux = NULL;//ayyi_mixer__aux_get_(ch->core_index, aux_num);
		int i; for(i=0;i<AYYI_AUX_PER_CHANNEL;i++){
			if(!chan->aux[i]) continue;
			aux = chan->aux[i];
			break;
		}
		if(aux)
			on_have_aux(t, NULL, NULL);
		else{
			am_aux__add(ch, 0, on_have_aux, NULL);
		}
	}

	module__ensure_audio_track(on_have_track, NULL);
}


void
module__reset_track(AyyiFinish on_finish, AMTrack* track, gpointer _user_data)
{
	//currently just disables solo.

	START_MODULE;

	void module_on_unsolo(AyyiIdent t, GError** error, gpointer user_data)
	{
		PF;
		FAIL_MODULE_IF_ERROR;
		AMTrack* track = user_data;
		assert(track, "track");
		assert(!am_track__is_solod(track), "track solod");

		FINISH_MODULE(t);
	}

	am_track__solo(track, FALSE, module_on_unsolo, track);
}


void
module__clear_auxs(AyyiFinish on_finish, AMTrack* track, gpointer _user_data)
{
	START_MODULE;

	typedef struct {
		Channel* ch;
		int      n_outstanding;
	} C;
	C* c = g_new0(C, 1);

	void module_clear_aux_on_clear(AyyiIdent t, GError** error, gpointer _c)
	{
		PF0;
		C* c = _c;
		FAIL_MODULE_IF_ERROR;
		int n_auxs = ayyi_mixer__aux_count(c->ch->core_index);
		assert(n_auxs > -1, "shm error");
		c->n_outstanding ++;
		dbg(0, "n_auxs=%i", n_auxs);
		if(!n_auxs && !c->n_outstanding){
			g_free(c);
			FINISH_MODULE(t);
		}
	}
	dbg(0, "callback=%p", module_clear_aux_on_clear);

	c->ch = am_track__lookup_channel(track);

	int i; for(i=0;i<AYYI_AUX_PER_CHANNEL;i++){
		c->n_outstanding ++;
		am_aux__remove(c->ch, i, module_clear_aux_on_clear, c);
	}
}



